/*
 * Copyright (c) 2024.
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.common.vorgang.GrpcServiceKontoTestFactory;
import de.ozgcloud.nachrichten.common.vorgang.Vorgang;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import io.grpc.stub.StreamObserver;

class AntragraumGrpcServiceTest {

	@Spy
	@InjectMocks
	private AntragraumGrpcService grpcService;
	@Mock
	private AntragraumService service;
	@Mock
	private AntragraumNachrichtMapper nachrichtMapper;
	@Mock
	private RueckfrageMapper rueckfrageMapper;

	@Mock
	private VorgangService vorgangService;

	@DisplayName("Find rueckfragen")
	@Nested
	class TestFindRueckfragen {

		@Captor
		private ArgumentCaptor<GrpcFindRueckfragenResponse> responseCaptor;
		@Mock
		private StreamObserver<GrpcFindRueckfragenResponse> streamObserver;

		private final RueckfrageHead rueckfrage = RueckfrageHeadTestFactory.create();

		private final GrpcRueckfrageHead grpcRueckfrageHead = GrpcRueckfrageHeadTestFactory.create();

		@BeforeEach
		void setup() {
			when(service.findRueckfragen(any())).thenReturn(Stream.of(rueckfrage));
			when(nachrichtMapper.toGrpc(any())).thenReturn(grpcRueckfrageHead);
		}

		@Test
		void shouldCallMapper() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(nachrichtMapper).toGrpc(rueckfrage);
		}

		@Test
		void shouldCallOnNext() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onNext(responseCaptor.capture());
			assertThat(responseCaptor.getValue().getRueckfrageHeadList()).containsExactly(grpcRueckfrageHead);
		}

		@Test
		void shouldCallOnCompleted() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onCompleted();
		}
	}

	@DisplayName("Send rueckfrage answer")
	@Nested
	class TestSendRueckfrageAnswer {
		@Mock
		private StreamObserver<GrpcSendRueckfrageAnswerResponse> streamObserver;
		private PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private RueckfrageHead rueckfrage = RueckfrageHeadTestFactory.create();

		@BeforeEach
		void setup() {
			when(nachrichtMapper.fromRueckfrageAnswer(any(GrpcRueckfrageAnswer.class))).thenReturn(postfachNachricht);
			when(service.sendRueckfrageAnswer(anyString(), anyString(), any(PostfachNachricht.class)))
					.thenReturn(UUID.randomUUID().toString());
			doReturn(postfachNachricht).when(grpcService).enrichRueckfrageAnswer(any(), any());
			doReturn(rueckfrage).when(grpcService).getRueckfrageByNachricht(anyString(), anyString());
		}

		@Test
		void shouldCallMapper() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(nachrichtMapper).fromRueckfrageAnswer(any(GrpcRueckfrageAnswer.class));
		}

		@Test
		void shouldGetRueckfrage() {
			sendRueckfrageAnswer();

			verify(grpcService).getRueckfrageByNachricht(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN,
					GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID);
		}

		@Test
		void shouldEnrichPostfachNachricht() {
			sendRueckfrageAnswer();

			verify(grpcService).enrichRueckfrageAnswer(postfachNachricht, rueckfrage);
		}

		@Test
		void shouldCallAntragraumService() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(service).sendRueckfrageAnswer(anyString(), anyString(), any(PostfachNachricht.class));
		}

		@Test
		void shouldCallOnNext() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onNext(any(GrpcSendRueckfrageAnswerResponse.class));
		}

		@Test
		void shouldCallOnCompleted() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onCompleted();
		}

		private void sendRueckfrageAnswer() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);
		}
	}

	@DisplayName("Get rueckfrage by nachricht")
	@Nested
	class TestGetRueckfrageByNachricht {

		private final RueckfrageHead matchingRueckfrage = RueckfrageHeadTestFactory.createBuilder().id(GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID)
				.build();
		private final RueckfrageHead rueckfrage = RueckfrageHeadTestFactory.create();

		@BeforeEach
		void mock() {
			when(service.findRueckfragen(any())).thenReturn(Stream.of(rueckfrage, matchingRueckfrage));
		}

		@Test
		void shouldCallService() {
			getRueckfrage();

			verify(service).findRueckfragen(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldReturnMatchingPostfachNachricht() {
			var rueckfrage = getRueckfrage();

			assertThat(rueckfrage).isEqualTo(matchingRueckfrage);
		}

		@Test
		void shouldThrowExceptionIfNoRueckfrageExists() {
			when(service.findRueckfragen(any())).thenReturn(Stream.empty());

			assertThatThrownBy(() -> getRueckfrage()).isInstanceOf(TechnicalException.class);
		}

		private RueckfrageHead getRueckfrage() {
			return grpcService.getRueckfrageByNachricht(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN,
					GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID);
		}
	}

	@DisplayName("Enrich postfachNachricht")
	@Nested
	class TestEnrichPostfachNachricht {

		private PostfachNachricht postfachNachrichtToEnrich = PostfachNachrichtTestFactory.createBuilder()
				.postfachAddress(null)
				.vorgangId(null)
				.sentAt(null)
				.build();

		private final RueckfrageHead postfachNachrichtRueckfrage = RueckfrageHeadTestFactory.create();

		@Test
		void shouldSetVorgangId() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getVorgangId()).isEqualTo(postfachNachrichtRueckfrage.getVorgangId());
		}

		@Test
		void shouldSetPostfachAddresses() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getPostfachAddress()).isEqualTo(postfachNachrichtRueckfrage.getPostfachAddress());
		}

		@Test
		void shouldSetSentAt() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getSentAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}
	}

	@DisplayName("Get rueckfrage")
	@Nested
	class TestGetRueckfrage {

		@Mock
		private StreamObserver<GrpcGetRueckfrageResponse> streamObserver;

		private final GrpcGetRueckfrageRequest request = GrpcGetRueckfrageRequestTestFactory.create();

		private final PostfachNachricht rueckfrage = PostfachNachrichtTestFactory.create();

		private final GrpcGetRueckfrageResponse getRueckfrageResponse = GrpcGetRueckfrageResponseTestFactory.create();

		private final GrpcRueckfrage grpcRueckfrage = GrpcRueckfrageTestFactory.create();

		@BeforeEach
		void mock() {
			when(service.getRueckfrage(any(), any())).thenReturn(rueckfrage);

			doReturn(getRueckfrageResponse).when(grpcService).buildGetRueckfrageResponse(any());
			doReturn(grpcRueckfrage).when(grpcService).enhanceRueckfrage(any(), any());
		}

		@Test
		void shouldCallServiceToGetRueckfrage() {
			getRueckfrage();

			verify(service).getRueckfrage(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldEnhanceRueckfrage() {
			getRueckfrage();

			verify(grpcService).enhanceRueckfrage(rueckfrage, GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldCallBuildRueckfrageResponse() {
			getRueckfrage();

			verify(grpcService).buildGetRueckfrageResponse(grpcRueckfrage);
		}

		@Test
		void shouldCallOnNext() {
			getRueckfrage();

			verify(streamObserver).onNext(getRueckfrageResponse);
		}

		@Test
		void shouldCallOnCompleted() {
			getRueckfrage();

			verify(streamObserver).onCompleted();
		}

		private void getRueckfrage() {
			grpcService.getRueckfrage(request, streamObserver);
		}
	}

	@DisplayName("Enhance rueckfrage")
	@Nested
	class TestEnhanceRueckfrage {

		@Captor
		private ArgumentCaptor<Stream<PostfachNachricht>> answersStreamCaptor;

		@Captor
		private ArgumentCaptor<List<PostfachNachricht>> answersListCaptor;

		private final PostfachNachricht rueckfrage = PostfachNachrichtTestFactory.create();

		private final PostfachNachricht answer = PostfachNachrichtTestFactory.create();

		private final ZonedDateTime now = ZonedDateTime.now();
		private final Optional<ZonedDateTime> sentAt = Optional.of(now);

		private final Vorgang vorgang = VorgangTestFactory.create();

		private final GrpcRueckfrage grpcRueckfrage = GrpcRueckfrageTestFactory.create();

		@BeforeEach
		void mock() {
			when(service.findAnswers(any(), any())).thenReturn(Stream.of(answer));
			when(service.getFirstSentAt(any())).thenReturn(sentAt);
			when(vorgangService.getVorgang(any())).thenReturn(vorgang);
			when(rueckfrageMapper.toRueckfrage(any(), any(), any())).thenReturn(grpcRueckfrage);
			when(service.isAccessible(any(), any())).thenReturn(true);

			doReturn(grpcRueckfrage).when(grpcService).addAnswers(any(), any());
		}

		@Test
		void shouldCallVorgangService() {
			enhanceRueckfrage();

			verify(vorgangService).getVorgang(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallServiceToGetAnswers() {
			enhanceRueckfrage();

			verify(service).findAnswers(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldCallServiceToGetFirstSentAt() {
			enhanceRueckfrage();

			verify(service).getFirstSentAt(answersStreamCaptor.capture());
			assertThat(answersStreamCaptor.getValue().toList().get(0)).isEqualTo(answer);
		}

		@Test
		void shouldCallMapper() {
			enhanceRueckfrage();

			verify(rueckfrageMapper).toRueckfrage(sentAt, vorgang, rueckfrage);
		}

		@Test
		void shouldCallAddAnswers() {
			enhanceRueckfrage();

			verify(grpcService).addAnswers(eq(grpcRueckfrage), answersListCaptor.capture());
			assertThat(answersListCaptor.getValue().get(0)).isEqualTo(answer);
		}

		@Test
		void shouldCallIsAccessable() {
			enhanceRueckfrage();

			verify(service).isAccessible(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);
		}

		@Test
		void shouldSetAccessible() {
			var rueckfrage = enhanceRueckfrage();

			assertThat(rueckfrage.getAccessible()).isTrue();
		}

		@Test
		void shouldReturnRueckfrage() {
			var buildedRueckfrage = enhanceRueckfrage();

			assertThat(buildedRueckfrage).isEqualTo(grpcRueckfrage);
		}

		private GrpcRueckfrage enhanceRueckfrage() {
			return grpcService.enhanceRueckfrage(rueckfrage, GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);
		}
	}

	@DisplayName("Add answers")
	@Nested
	class TestAddAnswers {

		private final PostfachNachricht answer = PostfachNachrichtTestFactory.create();

		private final GrpcRueckfrage grpcRueckfrage = GrpcRueckfrageTestFactory.create();

		private final GrpcRueckfrageAnswer grpcRueckfrageAnwswer = GrpcRueckfrageAnswerTestFactory.create();

		@BeforeEach
		void mock() {
			when(nachrichtMapper.toRueckfrageAnswer(any())).thenReturn(grpcRueckfrageAnwswer);
		}

		@Test
		void shouldCallMapper() {
			grpcService.addAnswers(grpcRueckfrage, List.of(answer));

			verify(nachrichtMapper).toRueckfrageAnswer(answer);
		}

		@Test
		void shouldReturnRueckfrageWithAnswers() {
			var enhancedRueckfrage = grpcService.addAnswers(grpcRueckfrage, List.of(answer));

			assertThat(enhancedRueckfrage.getAnswersList().get(0)).isEqualTo(grpcRueckfrageAnwswer);
		}
	}

	@DisplayName("Build get rueckfrage response")
	@Nested
	class TestBuildGetRueckfrageResponse {

		private final GrpcRueckfrage grpcRueckfrage = GrpcRueckfrageTestFactory.create();

		@Test
		void shouldReturnResponseIncludingRueckfrage() {
			var rueckfrageResponse = grpcService.buildGetRueckfrageResponse(grpcRueckfrage);

			assertThat(rueckfrageResponse.getRueckfrage()).isEqualTo(grpcRueckfrage);
		}
	}
}